#include "Chi2FCN.h"

#include "TVectorD.h"
#include "TFile.h"

#include <Math/Util.h>

#include <cmath>
#include <iostream>
#include <cassert>

Chi2FCN::Chi2FCN(int nTemplates, int nDec):
    fNTemp(nTemplates),
    fNBinsObs(0),
    fNDec(nDec),
    invCovVb(0),
    invCovVc(0),
    theErrorDef(1),
    IsMatrixSet(false),
    IsDataSet(false),
    fDebug(false)
{}

Chi2FCN::~Chi2FCN()
{
    if( IsMatrixSet )
    {
        for(int i=0; i<fNBinsObs; i++)
            for(int j=0; j<fNTemp; j++)
                delete[] T[i][j];

        for(int i=0; i<fNBinsObs; i++)
            delete[] T[i];

        delete[] T;

        delete[] Nmeas;
    }

}

void Chi2FCN::SetObs(TH1F** sgnMC, TH1F* bkgMC, TH1F* theData){

    fNBinsObs += sgnMC[0] -> GetNbinsX();

    hSgnMC.push_back(sgnMC); 
    hBkgMC.push_back(bkgMC); 

    hData.push_back(theData);
}

void Chi2FCN::SetCovariance(TMatrixD* covVb, TMatrixD* covVc)
{
    // invert sgn covariance matrix
    if( covVb != NULL )
    {
        invCovVb = (TMatrixD*) covVb -> Clone();

        invCovVb -> Invert();
    }
    else
        std::cout << "\nNo sgn covariance matrix provided: corresponding penalty term will not be considered for minimization" << std::endl;

    // invert background covariance matrix
    if( covVc != NULL )
    {
        invCovVc = (TMatrixD*) covVc -> Clone();

        invCovVc -> Invert();
    }
    else
        std::cout << "\nNo background covariance matrix provided: corresponding penalty term will not be considered for minimization" << std::endl;
}

void Chi2FCN::SetMatrix()
{
    // signal template matrix
    T = new double**[fNBinsObs];

    for(int i=0; i<fNBinsObs; i++)
    {
        T[i] = new double*[fNTemp];

        for(int j=0; j<fNTemp; j++)
            T[i][j] = new double[fNDec];
    }

    // set T matrix
    int nFilled(0);
    for(int mc=0; mc<hSgnMC.size(); mc++)
    {
        int nBinsMC = hSgnMC[mc][0]->GetNbinsX();

        for(int bn=0; bn<fNTemp; bn++)
        {
            for(int bmc=0; bmc<nBinsMC; bmc++){
                T[ nFilled + bmc ][bn][0] = hSgnMC[mc][bn] -> GetBinContent( bmc+1 );
            }
        }

        nFilled += nBinsMC;
    }

    // set background array template
    Nbkg = new double[fNBinsObs];

    nFilled = 0;
    for(int d=0; d<hBkgMC.size(); d++)
    {
        int nBinsMC = hBkgMC[d] -> GetNbinsX();

        for(int bmc=0; bmc<nBinsMC; bmc++)
            Nbkg[ nFilled + bmc ] = hBkgMC[d] -> GetBinContent( bmc+1 );

        nFilled += nBinsMC;
    }

    // set also the array of data
    Nmeas = new double[fNBinsObs];

    nFilled = 0;
    for(int d=0; d<hData.size(); d++)
    {
        int nBinsData = hData[d] -> GetNbinsX();

        for(int bd=0; bd<nBinsData; bd++)
            Nmeas[ nFilled + bd ] = hData[d] -> GetBinContent( bd+1 );

        nFilled += nBinsData;
    }

    IsMatrixSet = true;

}

double Chi2FCN::operator()(const std::vector<double>& par) const {

    assert( par.size() == fNTemp + fNBinsObs && "Wrong number of parameters provided.");

    if( invCovVb != NULL )
        assert( (invCovVb -> GetNcols() == fNTemp && invCovVb -> GetNrows() == fNTemp) && "Bad sgn covariance matrix provided: wrong number of columns/rows."  );

    if( invCovVc != NULL )
        assert( (invCovVc -> GetNcols() == fNBinsObs && invCovVc -> GetNrows() == fNBinsObs) && "Bad background covariance matrix provided: wrong number of columns/rows."  );

    if( fDebug )
        std::cout << "\n ================= DEBUG: call to fcn function ==================== " << std::endl;

    if(!IsMatrixSet)
    {
        const_cast<Chi2FCN*>(this) -> SetMatrix();
    }

    // parameters of interest (POI) and nuisance parameters
    double b[fNTemp];// POI
    double c[fNBinsObs];// nuisance par

    std::copy(par.begin(), par.begin()+fNTemp, b);
    std::copy(par.begin()+fNTemp, par.begin()+fNTemp+fNBinsObs, c);

    // predicted events for each observable bin
    double Npred[fNBinsObs];

    // initialize
    for(int bo=0; bo<fNBinsObs; bo++)
        Npred[bo] = 0;

    // compute predicted events
    for(int bo=0; bo<fNBinsObs; bo++)
    {
        std::vector<double> vNpred;

        // predicted signal
        for(int bn=0; bn<fNTemp; bn++)
            vNpred.push_back( b[bn]*T[bo][bn][0] );

        // predicted background
        vNpred.push_back( c[bo]*Nbkg[bo] );

        ROOT::Math::KahanSum<double> k;
        k.Add(vNpred.begin(), vNpred.end());

        Npred[bo] = k.Result();
    }

    // debug
    for(int i=0; i<par.size() && fDebug; i++)
        std::cout << "par[" << i << "] = " << par[i] << std::endl;

    for(int i=0; i<fNTemp && fDebug; i++)
        std::cout << "par b" << i << " = " << b[i] << std::endl;

    for(int i=0; i<fNBinsObs && fDebug; i++)
        std::cout << "par c" << i << " = " << c[i] << std::endl;

    double chi2(0);
    std::vector<double> chi2_sum;

    for(int bo=0; bo<fNBinsObs; bo++)
    {
        if( Npred[bo] == 0 || Nmeas[bo] == 0 )
            continue;

        // Simple chi2
        //chi2_sum.push_back( pow(Npred[bo] - Nmeas[bo], 2)/Nmeas[bo] );

        // mu - n + n*log(n)
        chi2_sum.push_back( 2.*Npred[bo] );
        chi2_sum.push_back( -2.*Nmeas[bo] );
        chi2_sum.push_back( 2.*Nmeas[bo]*log(Nmeas[bo]/Npred[bo]) );
        
        if( fDebug )
        {
            std::streamsize ss = std::cout.precision();
            std::cout.precision(40);
            std::cout << "\nBin = " << bo+1 << std::endl;
            std::cout << "Npred = " << Npred[bo] << std::endl;
            std::cout << "Nmeas = " << Nmeas[bo] << std::endl;
            std::cout << "Npred - Nmeas = " << Npred[bo] - Nmeas[bo] << std::endl;
            std::cout << "Nmeas[bo]/Npred[bo] = " << (double)Nmeas[bo]/Npred[bo] << std::endl;
            std::cout << "log(Nmeas[bo]/Npred[bo]) = " << log(Nmeas[bo]/Npred[bo]) << std::endl;
            std::cout << "Nmeas[bo]*log(Nmeas[bo]/Npred[bo]) = " << Nmeas[bo]*log(Nmeas[bo]/Npred[bo]) << std::endl;
            std::cout << "Single term = " << Npred[bo] - Nmeas[bo] + Nmeas[bo]*log(Nmeas[bo]/Npred[bo]) << std::endl;
            std::cout << "Ratio = " << Nmeas[bo]/Npred[bo] << std::endl;
            std::cout.precision (ss);
        }
    }

    // compute the penalty term
    double penaltyTerm(0);

    // sgn cov term contribution
    if( invCovVb != NULL )
    {
        double sgnTerm(0);

        std::vector<double> db;
        for(int bn=0; bn<fNTemp; bn++)
        {
            // no contribution if bin has zero content
            if( b[bn] == 0 )
                db.push_back(0.);
            else
                db.push_back(1. - b[bn]);
        }

        double debTot(0);
        for(int i=0; i<fNTemp; i++)
            for(int j=0; j<fNTemp; j++)
            {
                sgnTerm = db[i]*(*invCovVb)(i,j)*db[j];

                chi2_sum.push_back( sgnTerm );

                if( fDebug )
                {
                    std::cout << "sgn cov term b" << i << " = " << sgnTerm << std::endl;

                    debTot += sgnTerm;
                }
            }

        if( fDebug )
            std::cout << "Total sgn cov = " << debTot << std::endl;
    }

    // background cov term contribution
    if( invCovVc != NULL )
    {
        double bkgTerm(0);

        std::vector<double> dc;
        for(int bo=0; bo<fNBinsObs; bo++)
        {
            // no contribution if bin has zero content
            if( c[bo] == 0 )
                dc.push_back(0.);
            else
                dc.push_back(1. - c[bo]);
        }

        double debTot(0);
        for(int i=0; i<fNBinsObs; i++)
            for(int j=0; j<fNBinsObs; j++)
            {
                bkgTerm = dc[i]*(*invCovVc)(i,j)*dc[j];

                chi2_sum.push_back( bkgTerm );

                if( fDebug )
                {
                    std::cout << "Bkg cov term b" << i << " = " << bkgTerm << std::endl;

                    debTot += bkgTerm;
                }
            }

        if( fDebug )
            std::cout << "Total Bkg cov = " << debTot << std::endl;
    }

    ROOT::Math::KahanSum<double> kChi2;
    kChi2.Add(chi2_sum.begin(), chi2_sum.end());

    chi2 = kChi2.Result();

    if( fDebug )
        std::cout << "Chi2 = " << chi2 << std::endl;

    return chi2;
}
