// Test fit:
//
// g++ -g fitTest.C Chi2FCN.h Chi2FCN.C -o fitTest `root-config --cflags --glibs` -lMinuit2
//
#include "TH1F.h"
#include "TRandom.h"
#include "TF1.h"
#include "TCanvas.h"
#include "TROOT.h"
#include "TGraph.h"
#include "TH2F.h"
#include "THStack.h"
#include "TDecompChol.h"
#include "TStyle.h"
#include "TLegend.h"
#include "TStopwatch.h"

#include "Chi2FCN.h"

#include "Minuit2/MnUserParameters.h"
#include "Minuit2/MnMigrad.h"
#include "Minuit2/FunctionMinimum.h"
#include "Minuit2/MnPrint.h"
#include "Minuit2/MnHesse.h"
#include "Minuit2/MnUserParameterState.h"
#include "Minuit2/MnMinos.h"
#include "Minuit2/MnContours.h"
#include "Minuit2/MnEigen.h"

#include <vector>
#include <iostream>

using namespace std;

TVectorD getInitialWeights(int NBins, double sigma, TMatrixD* covM, string name)
{
    // simulate values for cov matrix
    double* covArr = new double[NBins*NBins];

    // simple covariance matrix (nearest 3 bins correlated) 
    for(int i=0; i<NBins; i++)
        for(int j=0; j<NBins; j++)
        {
            if( i == j )
                covArr[i*NBins + j] = 1.0;
            else if( j == i+1 || j == i-1 )
                covArr[i*NBins + j] = 0.2;
            else if( j == i+2 || j == i-2 )
                covArr[i*NBins + j] = 0.05;
            else
                covArr[i*NBins + j] = 0.;
        }

    covM -> SetMatrixArray(covArr);

    // starting from uncorrelated values of the b and c coefficients we can obtain correlated ones by 
    // applying the change of coordinates matrix obtained from the Cholesky decomposition of covariance matrix.
    // The obtained coefficients are trh true POI and nuisance
    //
    // Cholesky decomposition of covariance matrix (NOTE: actually this is the matrix divided by sigma^2)
    TDecompChol* ch = new TDecompChol(*covM);
    bool ret = ch -> Decompose();

    TMatrixD* U = new TMatrixD(NBins, NBins);
    *U = ch -> GetU();

    // weights
    double w[NBins];

    TRandom* rnd = new TRandom(0);
    for(int bn=0; bn<NBins; bn++)
        w[bn] = rnd->Gaus(0, sigma);

    TVectorD wDec(0, NBins-1, w);
    // The vector of correlated coefficient to be used to simulate data
    TVectorD wCor = *U*wDec;

    *covM *= pow(sigma, 2);

    covM -> Print();

    TCanvas* cFC = new TCanvas();
    cFC -> cd();
    gStyle->SetOptStat(0);

    covM -> Draw("colz");

    cFC -> SaveAs(Form("%s.pdf",name.c_str()));

    delete ch;
    delete U;
    delete rnd;
    delete cFC;

    return wCor;
}

int testSyst(bool debug=false)
{
    gROOT -> SetBatch(kTRUE);

    Minuit2::MnPrint::SetLevel(3);

    // Normalization factor for real data
    //
    // HIGH STATISTICS: seems to cause problems
    double fN = 6e11;
    
    // LOW STATISTICS: works fine
    //double fN = 57;


    // Generate templates and data
    const int nTemplates = 2;

    // number of bins and ranges for  observables
    const int NBinsO1 = 2; 
    const int MinO1 = 0;   
    const int MaxO1 = 400; 

    const int NBinsO2 = 2; 
    const int MinO2 = 0;   
    const int MaxO2 = 4400;;

    // Signal templates to fit the observables (will be produced from MC) 
    TH1F* h_sgn_obs1_mc[nTemplates];
    TH1F* h_sgn_obs2_mc[nTemplates];

    vector<int> templEvts = {1000, 2000};

    for(int b=0; b<nTemplates; b++)
    {
        h_sgn_obs1_mc[b] = new TH1F(Form("h_sgn_obs1_mc_t%d",b+1),"", NBinsO1, MinO1, MaxO1);
        h_sgn_obs2_mc[b] = new TH1F(Form("h_sgn_obs2_mc_t%d",b+1),"", NBinsO2, MinO2, MaxO2);
    }

    double energy, l;

    // produce templates from gaussian distributions
    TRandom rnd(0);
    for(int b=0; b<nTemplates; b++)
    {
        for(int n=0; n<templEvts[b]; n++)
        {
            energy = rnd.Gaus(20*(b+1),200);
            l = rnd.Gaus(2000+700*(b+1),1500);

            h_sgn_obs1_mc[b] -> Fill(energy);
            h_sgn_obs2_mc[b] -> Fill(l);
        }
    }

    // Background templates to fit the observables (will be produced from MC) 
    TH1F* h_bkg_obs1_mc = new TH1F("h_bkg_obs1_mc","", NBinsO1, MinO1, MaxO1);
    TH1F* h_bkg_obs2_mc = new TH1F("h_bkg_obs2_mc","", NBinsO2, MinO2, MaxO2);

    // produce templates from gaussian distributions
    for(int n=0; n<3500; n++)
    {
        energy = rnd.Gaus(50,100);
        l = rnd.Gaus(1600,1500);

        h_bkg_obs1_mc -> Fill(energy);
        h_bkg_obs2_mc -> Fill(l);
    }

    std::cout << "\n=============== Signal (b-coefficients - POI true values) Covariance Matrix ===================" << std::endl;
    // generic value of sigma
    double sigma = 0.2;

    // signal covariance matrix
    TMatrixD* sgnCov = new TMatrixD(nTemplates, nTemplates);

    TVectorD bCor = getInitialWeights(nTemplates, sigma, sgnCov, "sgn_covariance_matrix");

    std::cout << "\n=============== Background (c-coefficients - nuisance true values) Covariance Matrix ===================" << std::endl;
    int NBinsO = NBinsO1 + NBinsO2;

    // background covariance matrix
    TMatrixD* bkgCov = new TMatrixD(NBinsO, NBinsO);

    // reset sigma for background nuisances (now the same)
    sigma = 0.2;
    TVectorD cCor = getInitialWeights(NBinsO, sigma, bkgCov, "bkg_covariance_matrix");

    // RESET: fixed coeff values for test
    //bCor[0] =  1. - 1.2658    ;  
    //bCor[1] =  1. - 1.00516   ;
    //cCor[0] =  1. - 1.08225   ;
    //cCor[1] =  1. - 1.34759   ;
    //cCor[2] =  1. - 0.980988  ; 
    //cCor[3] =  1. - 1.1365    ;

    std::cout << "\n=============== Initial true values for b & c normalization coefficients ===================" << std::endl;
    for(int bn=0; bn<nTemplates; bn++)
        cout << "b" << bn << " = " << 1 + bCor[bn] << endl;

    for(int ci=0; ci<cCor.GetNoElements(); ci++)
        cout << "c" << ci << " = " << 1 + cCor[ci] << endl;

    // Generate data from:
    //
    // -> signal events by adding the different bin templates, and weighting each template with the true values of 
    //    the corresponding b-coefficient;
    // -> background events by adding the background template where each bin is weighted with the true values of 
    //    the corresponding c-coefficient;
    //
    // histos for simulated observables from Data (signal + background)
    TH1F* h_obs1_data = new TH1F("h_obs1_data","", NBinsO1, MinO1, MaxO1);
    TH1F* h_obs2_data = new TH1F("h_obs2_data","", NBinsO2, MinO2, MaxO2);

    double b(0), c(0);
    double sgn(0), bkg(0), obs(0);
    for(int bo=0; bo<NBinsO1; bo++)
    {
        // add signal contribution
        sgn = 0;
        for(int bn=0; bn<nTemplates; bn++)
        {
            b = 1 + bCor[bn];

            sgn += b*(h_sgn_obs1_mc[bn] -> GetBinContent(bo+1));
        }

        // add background contribution
        c = 1 + cCor[bo];

        bkg = c*(h_bkg_obs1_mc -> GetBinContent(bo+1));

        // set data normalization 
        obs = fN*(sgn + bkg);

        h_obs1_data -> SetBinContent(bo+1, obs);
    }

    for(int bo=0; bo<NBinsO2; bo++)
    {
        // add signal contribution
        sgn = 0;
        for(int bn=0; bn<nTemplates; bn++)
        {
            b = 1 + bCor[bn];

            sgn += b*(h_sgn_obs2_mc[bn] -> GetBinContent(bo+1));
        }

        // add background contribution
        c = 1 + cCor[bo + NBinsO1];// REMARK: need to get the right weights for current observable

        bkg = c*(h_bkg_obs2_mc -> GetBinContent(bo+1));

        // set data normalization
        obs = fN*(sgn + bkg);

        h_obs2_data -> SetBinContent(bo+1, obs);
    }

    // need to take into account the statistical error for each bin
    for(int b=1; b<NBinsO1; b++)
        h_obs1_data -> SetBinContent(b, rnd.PoissonD(h_obs1_data -> GetBinContent(b)));

    for(int b=1; b<NBinsO2; b++)
        h_obs2_data -> SetBinContent(b, rnd.PoissonD(h_obs2_data -> GetBinContent(b)));

    // set proper errors for data
    for(int b=1; b<NBinsO1; b++)
        h_obs1_data -> SetBinError(b, sqrt(h_obs1_data -> GetBinContent(b)));

    for(int b=1; b<NBinsO2; b++)
        h_obs2_data -> SetBinError(b, sqrt(h_obs2_data -> GetBinContent(b)));

    gStyle->SetOptStat(1);
    gStyle->SetOptStat("nmri");

    // Plot initial distributions
    TCanvas* c2 = new TCanvas("c2", "", 900, 1000);
    c2 -> Divide(2, nTemplates);

    for(int b=0; b<nTemplates; b++)
    {
        c2 -> cd(b*2+1);

        h_sgn_obs1_mc[b] -> GetXaxis() -> SetTitle("E");
        h_sgn_obs1_mc[b] -> GetYaxis() -> SetTitle("# evts");

        h_sgn_obs1_mc[b] -> Draw();

        c2 -> cd(b*2+2);

        h_sgn_obs2_mc[b] -> GetXaxis() -> SetTitle("L");
        h_sgn_obs2_mc[b] -> GetYaxis() -> SetTitle("# evts");

        h_sgn_obs2_mc[b] -> Draw();
    }

    c2 -> SaveAs("mc_observables_sgn_templates.pdf");

    TCanvas* c2b = new TCanvas();
    c2b -> Divide(2);

    c2b -> cd(1);

    h_bkg_obs1_mc -> GetXaxis() -> SetTitle("E");
    h_bkg_obs1_mc -> GetYaxis() -> SetTitle("# evts");

    h_bkg_obs1_mc -> Draw();

    c2b -> cd(2);

    h_bkg_obs2_mc -> GetXaxis() -> SetTitle("L");
    h_bkg_obs2_mc -> GetYaxis() -> SetTitle("# evts");

    h_bkg_obs2_mc -> Draw();

    c2b -> SaveAs("mc_observables_bkg_template.pdf");

    // Need to scale templates to match total normalization
    // of expected data
    for(int bn=0; bn<nTemplates; bn++)
    {
        h_sgn_obs1_mc[bn] -> Scale(fN);
        h_sgn_obs2_mc[bn] -> Scale(fN);
    }

    h_bkg_obs1_mc -> Scale(fN);
    h_bkg_obs2_mc -> Scale(fN);

    // We are ready for minimization 
    //
    // create FCN
    Chi2FCN myFCN(nTemplates,1);

    // set the observables: templates and data
    myFCN.SetObs(h_sgn_obs1_mc, h_bkg_obs1_mc, h_obs1_data);
    myFCN.SetObs(h_sgn_obs2_mc, h_bkg_obs2_mc, h_obs2_data);

    // set the covariance matrix
    myFCN.SetCovariance(sgnCov, bkgCov);

    // to get debug output from FCN
    if( debug )
        myFCN.SetDebug();

    // create Minuit parameters with names
    MnUserParameters upar;

    // initialize paramenters of interest
    for(int p=0; p<nTemplates; p++)
        upar.Add(Form("b%d",p), 1.1, 0.2, 0.0, 3);

    // initialize nuisance paramenters
    for(int p=nTemplates; p<nTemplates+NBinsO; p++)
        upar.Add(Form("c%d",p-nTemplates), 1.1, 0.2, 0.0, 10);

    // create MIGRAD minimizer
    MnMigrad migrad(myFCN, upar, 1);// constructor migrad(fcn, params, strategy [2=high-quality])

    // minimize
    FunctionMinimum min = migrad(1e6,0.001);//call migrad(ncalls, tolerance)
    //FunctionMinimum min = migrad(1e6,1);//call migrad(ncalls, tolerance)

    // output
    std::cout << "\n============= OUTPUT FROM MIGRAD STEP ==============" << std::endl;

    std::cout << min << std::endl;

    // save chi2 value at minimum for later
    double chi2MinValue = myFCN(min.UserState().Params());

    // call Hesse: 
    //
    // see 7.2.2 of https://root.cern.ch/download/minuit.pdf
    // (and end of that section)
    MnHesse mhesse(1);

    // get parameters from minimum
    MnUserParameterState parState = min.UserState();

    MnUserParameterState mHPar = mhesse(myFCN, parState, 1e6);// call mhesse(fcn, parState, max-calls) 

    std::cout << "============= OUTPUT FROM HESSE STEP ==============" << std::endl;

    std::cout << mHPar <<  std::endl;

    if( mHPar.CovarianceStatus() )
    {
        MnEigen meigv;
        vector<double> eigv = meigv(mHPar.Covariance());
        for(int e=0; e<eigv.size(); e++)
            std::cout << "Eigenvalue-" << e << " = " << eigv[e] << std::endl;
        std::cout<<std::endl;
    }

    std::cout << "============= DRAW FITTED OBSERVAABLES ==============" << std::endl;

    // parameter values and errors 
    std::vector<double> parVal = mHPar.Params();
    std::vector<double> parErr = mHPar.Errors();// errors from HESSE

    gStyle->SetEndErrorSize(5);
    gStyle->SetErrorX(0.);

    // fitted observables

    // E-obs1
    vector<double> vEfit(NBinsO1);

    TCanvas* cE = new  TCanvas();
    cE -> cd();

    TLegend* lE = new TLegend(0.8,0.4,0.9,0.75);
    lE -> AddEntry(h_obs1_data,"Data","p");

    THStack* hsE = new THStack();
    // add fitted background template
    TH1F* hEbkg = (TH1F*) h_bkg_obs1_mc -> Clone("h_bkg_obs1_weight");
    double wbkg(0);
    for(int bo=0; bo<NBinsO1; bo++)
    {
        wbkg = parVal[bo+nTemplates]*hEbkg -> GetBinContent(bo+1);
        hEbkg -> SetBinContent(bo+1, wbkg);

        vEfit[bo] = wbkg;
    }

    hEbkg -> SetFillStyle(3002);
    hEbkg -> SetFillColor(1);

    hsE -> Add(hEbkg);

    // add fitted signal templates
    TH1F* hE[nTemplates];
    for(int bn=0; bn<nTemplates; bn++)
    {
        hE[bn] = (TH1F*) h_sgn_obs1_mc[bn] -> Clone(Form("h_sgn_obs1_mc_t%d",bn+1));

        hE[bn] -> SetFillColor(bn+2);

        hE[bn] -> Scale(parVal[bn]);

        for(int bo=0; bo<NBinsO1; bo++)
            vEfit[bo] += parVal[bn]*h_sgn_obs1_mc[bn] -> GetBinContent(bo+1);

        lE -> AddEntry(hE[bn],Form("sgn tmpl %d",bn+1),"f");

        hsE -> Add(hE[bn]);
    }

    lE -> AddEntry(hEbkg,"Bkg","f");

    // draw stacked fitted templates
    hsE -> Draw("HIST");

    lE -> Draw();

    hsE -> GetXaxis() -> SetTitle("E");
    hsE -> GetYaxis() -> SetTitle("# evts");

    // draw data
    h_obs1_data -> SetMarkerStyle(8);
    h_obs1_data -> Draw("pe1same");

    // residuals
    double resE;
    for(int bo=0; bo<NBinsO1; bo++)
    {
        resE = (h_obs1_data -> GetBinContent(bo+1) - vEfit[bo]);
        std::cout << "E bo = " << bo+1 <<" - data = " << h_obs1_data -> GetBinContent(bo+1) << " - fit = " << vEfit[bo] << " - res = " << resE << std::endl;
    }

    cE -> SaveAs("fitted_obs1.pdf");

    // L-obs2 
    vector<double> vLfit(NBinsO2);

    TCanvas* cLE = new  TCanvas();
    cLE -> cd();

    TLegend* lL = new TLegend(0.8,0.4,0.9,0.75);
    lL -> AddEntry(h_obs2_data,"Data","p");

    THStack* hsL = new THStack();
    // add fitted background template
    TH1F* hLbkg = (TH1F*) h_bkg_obs2_mc -> Clone("h_bkg_obs2_weight");
    for(int bo=0; bo<NBinsO2; bo++)
    {
        wbkg = parVal[bo+nTemplates+NBinsO1]*hLbkg -> GetBinContent(bo+1);
        hLbkg -> SetBinContent(bo+1, wbkg);

        vLfit[bo] = wbkg;
    }

    hLbkg -> SetFillStyle(3002);
    hLbkg -> SetFillColor(1);

    hsL -> Add(hLbkg);

    // add fitted signal templates
    TH1F* hL[nTemplates];
    for(int bn=0; bn<nTemplates; bn++)
    {
        hL[bn] = (TH1F*) h_sgn_obs2_mc[bn] -> Clone(Form("h_sgn_obs2_mc_t%d",bn+1));

        hL[bn] -> SetFillColor(bn+2);

        hL[bn] -> Scale(parVal[bn]);

        for(int bo=0; bo<NBinsO2; bo++)
            vLfit[bo] += parVal[bn]*h_sgn_obs2_mc[bn] -> GetBinContent(bo+1);

        lL -> AddEntry(hL[bn],Form("sgn tmpl %d",bn+1),"f");

        hsL -> Add(hL[bn]);
    }

    lL -> AddEntry(hLbkg,"Bkg","f");

    // draw stacked fitted templates
    hsL -> Draw("HIST");

    lL -> Draw();

    hsL -> GetXaxis() -> SetTitle("L");
    hsL -> GetYaxis() -> SetTitle("# evts");

    // draw data
    h_obs2_data -> SetMarkerStyle(8);
    h_obs2_data -> Draw("pe1same");

    // residuals
    double resL;
    for(int bo=0; bo<NBinsO2; bo++)
    {
        resL = (h_obs2_data -> GetBinContent(bo+1) - vLfit[bo]);
        std::cout << "L bo = " << bo+1 <<" - data = " << h_obs2_data -> GetBinContent(bo+1) << " - fit = " << vLfit[bo] << " - res = " << resL << std::endl;
    }

    cLE -> SaveAs("fitted_obs2.pdf");

    std::cout << "\n============= START MINOS ANALYSIS ==============" << std::endl;
    // create MINOS error factory
    MnMinos minos(myFCN, min, 1);

    // 1-sigma MINOS errors
    std::map<int, std::pair<double,double> > eb;

    for(int p=0; p<nTemplates; p++)
        eb[ p ] = minos(p, 1e5, 0.001);// call mminos(par, maxcalls, tolerance) (see MnMino.h)

    std::cout << "\n============= OUTPUT FROM MINOS STEP ==============" << std::endl;

    // output
    std::cout<<"\n1-sigma minos errors: "<<std::endl;
    for(int p=0; p<nTemplates; p++)
        std::cout << Form("par b%d: ",p)  << min.UserState().Value(p) << " " << eb[p].first << " " << eb[p].second << std::endl;

}

int main(int argc, char** argv)
{

    if( argc != 1 && argc != 2)
    {
        cout << "\nUsage: \n\n ./fitTest debug \n\nDebug flag on: debug = 1\n" << endl;

        return 1;
    }

    bool debug = 0;

    if( argc == 3 )
        debug = atoi(argv[1]);

    testSyst(debug);

    return 0;
}
